# Copyright (c) 2023-2024, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
import torch

import pylibcugraphops

import e3nn.o3 as o3
from e3nn.o3 import Irreps
from e3nn.o3._reduce import _wigner_nj

from typing import Iterator, List

import pylibcugraphops.operators as ops


def _transpose_irrep_to_m_last(tensor: torch.Tensor, irreps: Irreps):
    new_tensor = torch.empty_like(tensor)
    offset = 0
    for irr in irreps:
        num_channels_per_m = irr.mul
        m = irr.ir.dim
        tmp = tensor[:, offset : offset + irr.dim].clone()
        if m > 1:
            tmp = tmp.reshape(-1, m, num_channels_per_m)
            tmp = torch.transpose(tmp, 2, 1).reshape(-1, irr.dim)
        new_tensor[:, offset : offset + irr.dim] = tmp
        offset += irr.dim

    return new_tensor


def _transpose_irrep_to_channels_last(tensor: torch.Tensor, irreps: Irreps):
    new_tensor = torch.empty_like(tensor)
    offset = 0
    for irr in irreps:
        num_channels_per_m = irr.mul
        m = irr.ir.dim
        tmp = tensor[:, offset : offset + irr.dim].clone()
        if m > 1:
            tmp = tmp.reshape(-1, num_channels_per_m, m)
            tmp = torch.transpose(tmp, 2, 1).reshape(-1, irr.dim)
        new_tensor[:, offset : offset + irr.dim] = tmp
        offset += irr.dim

    return new_tensor


def _irreps_to_segment_info(irreps: Irreps):
    info = []
    offset = 0
    is_sh = True
    for irr in irreps:
        info.append([offset, irr.ir.dim, irr.mul, -1])
        offset += irr.dim
        is_sh = is_sh and (irr.mul == 1)

    if is_sh:
        return None
    return torch.IntTensor(info)


class _transpose_irrep_to_channels_last_autograd(torch.autograd.Function):
    @staticmethod
    def forward(ctx, tensor: torch.Tensor, segment_info: torch.Tensor):
        segment_info = segment_info.detach().contiguous()
        ctx.save_for_backward(segment_info)
        tensor_transpose = torch.empty_like(tensor)
        stream = torch.cuda.current_stream().cuda_stream

        if tensor.numel() > 0:
            ops.segmented_transpose(
                tensor_transpose.detach().contiguous(),
                tensor.detach().contiguous(),
                segment_info,
                True,
                stream,
            )
        return tensor_transpose

    @staticmethod
    def backward(ctx, grad_out: torch.Tensor):
        (segment_info,) = ctx.saved_tensors
        grad_tensor = _transpose_irrep_to_m_last_autograd.apply(grad_out, segment_info)
        return grad_tensor, None


class _transpose_irrep_to_m_last_autograd(torch.autograd.Function):
    @staticmethod
    def forward(ctx, tensor: torch.Tensor, segment_info: torch.Tensor):
        segment_info = segment_info.detach().contiguous()
        ctx.save_for_backward(segment_info)
        tensor_transpose = torch.empty_like(tensor)
        stream = torch.cuda.current_stream().cuda_stream

        if tensor.numel() > 0:
            ops.segmented_transpose(
                tensor_transpose.detach().contiguous(),
                tensor.detach().contiguous(),
                segment_info,
                False,
                stream,
            )
        return tensor_transpose

    @staticmethod
    def backward(ctx, grad_out: torch.Tensor):
        (segment_info,) = ctx.saved_tensors
        grad_tensor = _transpose_irrep_to_channels_last_autograd.apply(
            grad_out, segment_info
        )
        return grad_tensor, None


def transpose_irrep_to_channels_last(
    tensor: torch.Tensor, segment_info: torch.Tensor
) -> torch.Tensor:
    if segment_info is None:
        return tensor
    return _transpose_irrep_to_channels_last_autograd.apply(tensor, segment_info)


def transpose_irrep_to_m_last(
    tensor: torch.Tensor, segment_info: torch.Tensor
) -> torch.Tensor:
    if segment_info is None:
        return tensor
    return _transpose_irrep_to_m_last_autograd.apply(tensor, segment_info)


def sort_tp_info_for_operand(
    path_offsets_and_dims: torch.Tensor, path_cg_values: torch.Tensor, op_idx: int
):
    # idx = 0 -> in1
    # idx = 1 -> in2
    # idx = 2 -> weight
    # idx = 3 -> out
    key = path_offsets_and_dims[:, op_idx]
    perm = key.sort(descending=False, stable=True)[1]
    return path_offsets_and_dims[perm, :], path_cg_values[perm]


def path_offsets_to_csr(
    path_offsets_and_dims: torch.Tensor, op_idx: int, op_stride: int
):
    diff = torch.diff(
        path_offsets_and_dims[:, op_idx],
        n=1,
        append=path_offsets_and_dims.new_tensor([op_stride]),
    )
    path_csr_offsets = diff.nonzero().to(dtype=torch.int32).squeeze(dim=1)
    path_csr_offsets = torch.cat(
        [
            path_csr_offsets.new_tensor([0]),
            path_csr_offsets + 1,
        ]
    )
    return path_csr_offsets


def make_tp_info(
    op_idx: int,
    op_stride: int,
    path_offsets: torch.Tensor,
    path_cg_values: torch.Tensor,
):
    path_offsets = path_offsets.clone().detach()
    path_cg_values = path_cg_values.clone().detach()
    path_offsets, path_cg_values = sort_tp_info_for_operand(
        path_offsets, path_cg_values, op_idx
    )
    path_csr_offsets = path_offsets_to_csr(path_offsets, op_idx, op_stride)

    path_cg_values = path_cg_values
    path_offsets = path_offsets
    path_csr_offsets = path_csr_offsets

    return (
        path_csr_offsets,
        path_offsets,
        path_cg_values,
    )


def get_tp_infos_from_e3nn_instructions(
    instructions: Iterator[o3.Instruction],
    irreps_in1: o3.Irreps,
    irreps_in2: o3.Irreps,
    irreps_out: o3.Irreps,
):
    connection_mode = instructions[0].connection_mode

    path_offsets_and_dims = []
    path_cg_values = []

    slices_a = irreps_in1.slices()
    slices_b = irreps_in2.slices()
    slices_o = irreps_out.slices()

    off_weight = 0

    for inst in instructions:
        error_msg = "Only one connection mode at a time supported"
        error_msg += f"got {connection_mode} and {inst.connection_mode} respectively"
        if not inst.connection_mode == connection_mode:
            raise ValueError(error_msg)
        ir_idx_a = inst.i_in1
        ir_idx_b = inst.i_in2
        ir_idx_o = inst.i_out

        off_a = slices_a[ir_idx_a].start
        off_b = slices_b[ir_idx_b].start
        off_o = slices_o[ir_idx_o].start
        path_weight = inst.path_weight

        dim_u = inst.path_shape[0]
        dim_v = inst.path_shape[1] if len(inst.path_shape) > 1 else 0
        dim_w = inst.path_shape[2] if len(inst.path_shape) > 2 else 0

        if connection_mode == "uvw":
            dim_weight = dim_u * dim_v * dim_w
        elif connection_mode in {"uvu", "uvv"}:
            dim_weight = dim_u * dim_v
        elif connection_mode == "uuw":
            dim_w = dim_v
            dim_v = dim_u
            dim_weight = dim_u * dim_w
        else:
            dim_weight = 0

        # dim_a, dim_b, dim_o here actually are the number of channels
        dim_a = irreps_in1[ir_idx_a].mul
        dim_b = irreps_in2[ir_idx_b].mul
        dim_o = irreps_out[ir_idx_o].mul

        l_a = irreps_in1[ir_idx_a].ir.l
        l_b = irreps_in2[ir_idx_b].ir.l
        l_o = irreps_out[ir_idx_o].ir.l

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            cg = o3.wigner_3j(l_a, l_b, l_o)

        for m_a in range(2 * l_a + 1):
            for m_b in range(2 * l_b + 1):
                for m_o in range(2 * l_o + 1):
                    cg_val = cg[m_a, m_b, m_o]
                    if cg_val != 0:
                        path_cg_values.append(cg_val * path_weight)
                        path_offsets_and_dims.append(
                            (
                                off_a + m_a * dim_a,
                                off_b + m_b * dim_b,
                                off_weight,
                                off_o + m_o * dim_o,
                                dim_u,
                                dim_v,
                                dim_w,
                                -1,  # place-holder / padding
                            )
                        )

        if inst.has_weight:
            off_weight += dim_weight

    strides = (
        irreps_in1.dim,
        irreps_in2.dim,
        off_weight,
        irreps_out.dim,
    )

    if connection_mode == "uvw" and all([ir.mul == 1 for ir in irreps_in1]):
        connection_mode = "1vw"
        pass
    elif connection_mode == "uvw" and all([ir.mul == 1 for ir in irreps_in2]):
        connection_mode = "u1w"
        pass
    connection_mode = getattr(pylibcugraphops.ConnectionMode, connection_mode.upper())

    path_cg_values = torch.FloatTensor(path_cg_values)
    path_offsets_and_dims = torch.IntTensor(path_offsets_and_dims)

    tp_info_fwd = make_tp_info(3, strides[3], path_offsets_and_dims, path_cg_values)
    tp_info_bwd_dgrad_in1 = make_tp_info(
        0, strides[0], path_offsets_and_dims, path_cg_values
    )
    tp_info_bwd_dgrad_in2 = make_tp_info(
        1, strides[1], path_offsets_and_dims, path_cg_values
    )
    tp_info_bwd_dgrad_w = make_tp_info(
        2, strides[2], path_offsets_and_dims, path_cg_values
    )

    return (
        tp_info_fwd,
        tp_info_bwd_dgrad_in1,
        tp_info_bwd_dgrad_in2,
        tp_info_bwd_dgrad_w,
        connection_mode,
        strides,
    )


def get_linear_infos_from_e3nn_instructions(
    instructions: Iterator[o3.Instruction],
    irreps_in: o3.Irreps,
    irreps_out: o3.Irreps,
):
    connection_mode = "uuvv"

    path_offsets_and_dims = []
    path_cg_values = []

    slices_i = irreps_in.slices()
    slices_o = irreps_out.slices()

    off_weight = 0

    for inst in instructions:
        ir_idx_i = inst.i_in
        ir_idx_o = inst.i_out

        off_i = slices_i[ir_idx_i].start
        off_o = slices_o[ir_idx_o].start
        path_weight = inst.path_weight

        dim_u = inst.path_shape[0]
        dim_v = inst.path_shape[1]

        ir_i = irreps_in[ir_idx_i].ir
        ir_o = irreps_out[ir_idx_o].ir
        l_i = ir_i.l
        l_o = ir_o.l
        assert l_i == l_o

        for m in range(2 * l_i + 1):
            path_cg_values.append(path_weight)
            path_offsets_and_dims.append(
                (
                    off_i + m * dim_u,
                    off_weight,
                    -1,
                    off_o + m * dim_v,
                    dim_u,
                    dim_v,
                    -1,
                    -1,  # place-holder / padding
                )
            )

        off_weight += dim_u * dim_v

    strides = (
        irreps_in.dim,
        off_weight,
        -1,
        irreps_out.dim,
    )

    connection_mode = getattr(pylibcugraphops.ConnectionMode, connection_mode.upper())

    path_cg_values = torch.FloatTensor(path_cg_values)
    path_offsets_and_dims = torch.IntTensor(path_offsets_and_dims)

    tp_info_fwd = make_tp_info(3, strides[3], path_offsets_and_dims, path_cg_values)
    tp_info_bwd_dgrad_in = make_tp_info(
        0, strides[0], path_offsets_and_dims, path_cg_values
    )
    tp_info_bwd_dgrad_w = make_tp_info(
        1, strides[1], path_offsets_and_dims, path_cg_values
    )

    return (
        tp_info_fwd,
        tp_info_bwd_dgrad_in,
        tp_info_bwd_dgrad_w,
        connection_mode,
        strides,
    )


# taken from https://github.com/ACEsuit/mace/blob/main/mace/tools/cg.py
def u_matrix_real(
    irreps_in: o3.Irreps,
    irreps_out: o3.Irreps,
    correlation: int,
    normalization: str = "component",
    filter_ir_mid=None,
    dtype=None,
):
    irrepss = [o3.Irreps(irreps_in)] * correlation
    if correlation == 4:
        filter_ir_mid = [
            (0, 1),
            (1, -1),
            (2, 1),
            (3, -1),
            (4, 1),
            (5, -1),
            (6, 1),
            (7, -1),
            (8, 1),
            (9, -1),
            (10, 1),
            (11, -1),
        ]
    wigners = _wigner_nj(
        *irrepss, normalization=normalization, filter_ir_mid=filter_ir_mid, dtype=dtype
    )
    current_ir = wigners[0][0]
    out = []
    stack = torch.tensor([])

    for ir, _, base_o3 in wigners:
        if ir in irreps_out and ir == current_ir:
            stack = torch.cat((stack, base_o3.squeeze().unsqueeze(-1)), dim=-1)
            last_ir = current_ir
        elif ir in irreps_out and ir != current_ir:
            if len(stack) != 0:
                out += [last_ir, stack]
            stack = base_o3.squeeze().unsqueeze(-1)
            current_ir, last_ir = ir, ir
        else:
            current_ir = ir
    out += [last_ir, stack]
    return out


def get_sym_cont_u_matrices(irreps_in, irreps_out, correlation):
    u_matrices = []
    coupling_irreps = o3.Irreps([irrep.ir for irrep in irreps_in])
    for ir in irreps_out:
        u_list = []
        for nu in range(0, correlation):
            u_matrix = u_matrix_real(
                irreps_in=coupling_irreps,
                irreps_out=ir,
                correlation=nu + 1,
            )[-1]
            if ir.ir.dim == 1:
                u_matrix = u_matrix.unsqueeze(0)
            u_list.append(u_matrix)
        u_matrices.append(u_list)

    return u_matrices


def init_sym_cont_weights(u_matrices, num_elements, num_features, correlation):
    weights = []
    for u_list in u_matrices:
        w_list = []
        for nu in range(0, correlation):
            num_params = u_list[nu].size(-1)
            w = torch.randn((num_elements, num_params, num_features)) / num_params
            w_list.append(w)
        weights.append(w_list)
    return weights


def _uniquify_index_tuple(tup):
    nnz, idx = torch.unique(torch.vstack(tup), dim=1, return_inverse=True)
    nnz_tup = tuple([t.squeeze(0) for t in torch.tensor_split(nnz, nnz.size(0))])
    return nnz_tup, nnz, idx


def _get_nonzero_indices(w):
    tmp = w.sum(dim=-1).sum(dim=-1).nonzero(as_tuple=True)
    return tmp, torch.vstack(tmp)


def contract_sym_cont_weights(
    weights: List[List[torch.Tensor]],
    u_matrices: List[List[torch.Tensor]],
    correlation: int,
):
    contracted_weights = []
    for nu in range(0, correlation):
        w_list = []
        for idx in range(len(u_matrices)):
            u = u_matrices[idx][nu]
            w = weights[idx][nu]
            u_shape = u.shape
            u = u.view(-1, u_shape[-1])
            tmp = torch.einsum("xb,abc->xca", u, w)
            tmp = tmp.view(*u_shape[:-1], tmp.size(-2), tmp.size(-1))
            w_list.append(tmp)
        tmp = torch.cat(w_list, dim=0)
        contracted_weights.append(tmp)

    return contracted_weights


def get_sym_contraction_path_offsets(
    weights, u_matrices, correlation, num_features, num_elements
):
    weight_indices = [None] * correlation
    tp_infos = [None] * correlation
    strides = [None] * correlation

    # contract weights as this makes selecting the indices easier
    with torch.no_grad():
        contracted_weights = contract_sym_cont_weights(weights, u_matrices, correlation)
    w_max = contracted_weights[-1]
    nnz_tup, _ = _get_nonzero_indices(w_max)

    for nu in range(correlation - 1, -1, -1):
        nnz_tup_next, _, rev_idx = _uniquify_index_tuple(nnz_tup[:-1])
        tmp = []

        if nu == correlation - 1:
            for n in range(nnz_tup[0].size(0)):
                tmp.append(
                    (
                        nnz_tup[-1][n].item() * num_features,
                        0,
                        n * num_features * num_elements,
                        rev_idx[n].item() * num_features,
                        num_features,
                        num_elements,
                        -1,
                        -1,
                    )
                )

            strides[nu] = (
                w_max.size(1) * num_features,
                num_elements,
                nnz_tup[0].size(0) * num_features * num_elements,
                nnz_tup_next[0].size(0) * num_features,
            )

            offset_and_dims = torch.IntTensor(tmp)
            cg_values = torch.ones(len(tmp), dtype=torch.float)

            tp_info_fwd = make_tp_info(3, strides[nu][3], offset_and_dims, cg_values)
            tp_info_bwd_dgrad_x = make_tp_info(
                0, strides[nu][0], offset_and_dims, cg_values
            )
            tp_info_bwd_dgrad_y = make_tp_info(
                1, strides[nu][1], offset_and_dims, cg_values
            )
            tp_info_bwd_dgrad_w = make_tp_info(
                2, strides[nu][2], offset_and_dims, cg_values
            )

            tp_infos[nu] = (
                tp_info_fwd,
                tp_info_bwd_dgrad_x,
                tp_info_bwd_dgrad_y,
                tp_info_bwd_dgrad_w,
            )

        else:
            for n in range(nnz_tup[0].size(0)):
                tmp.append(
                    (
                        nnz_tup[-1][n].item() * num_features,
                        n * num_features,
                        -1,
                        rev_idx[n].item() * num_features,
                        num_features,
                        -1,
                        -1,
                        -1,
                    )
                )

            strides[nu] = (
                w_max.size(1) * num_features,
                nnz_tup[0].size(0) * num_features,
                -1,
                nnz_tup_next[0].size(0) * num_features,
            )

            offset_and_dims = torch.IntTensor(tmp)
            cg_values = torch.ones(len(tmp), dtype=torch.float)

            tp_info_fwd = make_tp_info(3, strides[nu][3], offset_and_dims, cg_values)
            tp_info_bwd_dgrad_x = make_tp_info(
                0, strides[nu][0], offset_and_dims, cg_values
            )
            tp_info_bwd_dgrad_w = make_tp_info(
                1, strides[nu][1], offset_and_dims, cg_values
            )

            tp_infos[nu] = (tp_info_fwd, tp_info_bwd_dgrad_x, tp_info_bwd_dgrad_w)

        weight_indices[nu] = nnz_tup
        nnz_tup = nnz_tup_next

    return tp_infos, weight_indices, strides



# Copyright (c) 2023-2024, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
import torch
import torch.nn as nn
import e3nn.o3 as o3
import pylibcugraphops

from typing import Tuple, Optional, Iterator, List, Union

from pylibcugraphops.pytorch.utils import (
    maybe_detach,
    maybe_empty_like,
    get_operator_from_module,
)
import pylibcugraphops.operators as ops

from pylibcugraphops.pytorch.operators.fused_tensor_product_utils import (
    _irreps_to_segment_info,
    transpose_irrep_to_m_last,
    transpose_irrep_to_channels_last,
    get_tp_infos_from_e3nn_instructions,
    get_linear_infos_from_e3nn_instructions,
    get_sym_cont_u_matrices,
    init_sym_cont_weights,
    get_sym_contraction_path_offsets,
    contract_sym_cont_weights,
)


class TensorProductPathData(nn.Module):
    def __init__(
        self,
        path_csr_offsets: torch.Tensor,
        path_offsets: torch.Tensor,
        path_cg_values: torch.Tensor,
    ):
        super().__init__()
        # for now, don't have these buffers as part of the state dict to avoid
        # breaking changes w.r.t. loading state-dicts for now
        self.register_buffer("path_csr_offsets", path_csr_offsets, persistent=False)
        self.register_buffer("path_offsets", path_offsets, persistent=False)
        self.register_buffer("path_cg_values", path_cg_values, persistent=False)

    def forward(self, *args, **kwargs):
        raise NotImplementedError(
            "TensorProductPathData module is not intended to be called."
        )

    def c_type(self):
        return pylibcugraphops.make_tp_info(
            self.path_csr_offsets,
            self.path_offsets,
            self.path_cg_values,
        )


class _fused_tensor_product_autograd(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        in1: torch.Tensor,
        in2: torch.Tensor,
        weight: Optional[torch.Tensor],
        bias: Optional[torch.Tensor],
        args: Tuple,
    ) -> torch.Tensor:
        (
            tp_info_fwd,
            tp_info_bwd_dgrad_in1,
            tp_info_bwd_dgrad_in2,
            tp_info_bwd_dgrad_w,
            connection_mode,
            stride_out,
        ) = args

        out = torch.empty((in1.size(0), stride_out), device=in1.device, dtype=in1.dtype)
        stream = torch.cuda.current_stream().cuda_stream

        fwd_fun = get_operator_from_module(
            ops,
            "fused_tp_n2n_fwd",
            (
                in1.dtype,
                in2.dtype,
                weight.dtype if weight is not None else out.dtype,
                out.dtype,
            ),
        )
        fwd_fun(
            maybe_detach(out),
            maybe_detach(in1),
            maybe_detach(in2),
            maybe_detach(weight),
            maybe_detach(bias),
            connection_mode,
            tp_info_fwd,
            stream_id=stream,
        )

        ctx.save_for_backward(
            in1,
            in2,
            weight,
            bias,
        )
        ctx.args = args

        return out

    @staticmethod
    def backward(
        ctx, grad_out: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, None]:

        (in1, in2, weight, bias) = ctx.saved_tensors

        return _fused_tensor_product_backward_autograd.apply(
            grad_out,
            in1,
            in2,
            weight,
            bias,
            (*ctx.args, *ctx.needs_input_grad),
        )


class _fused_tensor_product_backward_autograd(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        grad_out: torch.Tensor,
        in1: torch.Tensor,
        in2: torch.Tensor,
        weight: Optional[torch.Tensor],
        bias: Optional[torch.Tensor],
        args,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, None, None]:
        (
            tp_info_fwd,
            tp_info_bwd_dgrad_in1,
            tp_info_bwd_dgrad_in2,
            tp_info_bwd_dgrad_w,
            connection_mode,
            stride_out,
            needs_input_grad_in1,
            needs_input_grad_in2,
            needs_input_grad_w,
            needs_input_grad_bias,
            needs_input_grad_args,
        ) = args

        # calling function takes care of ensuring contiguous and detach
        grad_in1 = maybe_empty_like(in1, needs_input_grad_in1)
        grad_in2 = maybe_empty_like(in2, needs_input_grad_in2)
        grad_weight = maybe_empty_like(weight, needs_input_grad_w)
        grad_bias = maybe_empty_like(bias, needs_input_grad_bias)

        stream = torch.cuda.current_stream().cuda_stream

        bwd_fun = get_operator_from_module(
            ops,
            "fused_tp_n2n_bwd",
            (
                in1.dtype,
                in2.dtype,
                weight.dtype if weight is not None else grad_out.dtype,
                grad_out.dtype,
            ),
        )

        bwd_fun(
            grad_in1,
            grad_in2,
            grad_weight,
            grad_bias,
            maybe_detach(grad_out),
            maybe_detach(in1),
            maybe_detach(in2),
            maybe_detach(weight),
            maybe_detach(bias),
            connection_mode,
            tp_info_bwd_dgrad_in1,
            tp_info_bwd_dgrad_in2,
            tp_info_bwd_dgrad_w,
            stream_id=stream,
        )

        ctx.args = args
        ctx.save_for_backward(
            in1,
            in2,
            weight,
            grad_out,
        )

        return grad_in1, grad_in2, grad_weight, grad_bias, None

    @staticmethod
    def backward(
        ctx, grad_grad_in1, grad_grad_in2, grad_grad_weight, grad_grad_bias, grad_args
    ):
        # args should not require_grad in any case
        # assert not ctx.needs_input_grad[4]
        in1, in2, weight, grad_out = ctx.saved_tensors

        grad_grad_out = maybe_empty_like(grad_out, ctx.needs_input_grad[0])
        grad_in1 = maybe_empty_like(in1, ctx.needs_input_grad[1])
        grad_in2 = maybe_empty_like(in2, ctx.needs_input_grad[2])
        grad_weight = maybe_empty_like(weight, ctx.needs_input_grad[3])

        stream = torch.cuda.current_stream().cuda_stream

        tp_info_fwd = ctx.args[0]
        tp_info_bwd_dgrad_in1 = ctx.args[1]
        tp_info_bwd_dgrad_in2 = ctx.args[2]
        tp_info_bwd_dgrad_w = ctx.args[3]
        connection_mode = ctx.args[4]

        bwd_bwd_fun = get_operator_from_module(
            ops,
            "fused_tp_n2n_bwd_bwd",
            (
                in1.dtype,
                in2.dtype,
                weight.dtype if weight is not None else grad_out.dtype,
                grad_out.dtype,
            ),
        )

        bwd_bwd_fun(
            grad_in1,
            grad_in2,
            grad_weight,
            grad_grad_out,
            maybe_detach(grad_grad_in1),
            maybe_detach(grad_grad_in2),
            maybe_detach(grad_grad_weight),
            maybe_detach(grad_out),
            maybe_detach(in1),
            maybe_detach(in2),
            maybe_detach(weight),
            connection_mode,
            tp_info_fwd,
            tp_info_bwd_dgrad_in1,
            tp_info_bwd_dgrad_in2,
            tp_info_bwd_dgrad_w,
            stream_id=stream,
        )

        if grad_grad_bias is not None:
            grad_grad_out += grad_grad_bias

        return grad_grad_out, grad_in1, grad_in2, grad_weight, None, None


class FusedTensorProduct(nn.Module):
    def __init__(
        self,
        instructions: Iterator[o3.Instruction],
        irreps_in1: o3.Irreps,
        irreps_in2: o3.Irreps,
        irreps_out: o3.Irreps,
        e3nn_compat_mode: bool = False,
        internal_weights: Optional[torch.Tensor] = None,
        shared_weights: Optional[bool] = False,
    ):
        super().__init__()

        tp_infos = get_tp_infos_from_e3nn_instructions(
            instructions,
            irreps_in1,
            irreps_in2,
            irreps_out,
        )

        self.e3nn_compat_mode = e3nn_compat_mode
        self.connection_mode = tp_infos[4]
        self.stride_in1 = tp_infos[5][0]
        self.stride_in2 = tp_infos[5][1]
        self.stride_w = tp_infos[5][2]
        self.stride_out = tp_infos[5][3]
        self.irreps_in1 = irreps_in1
        self.irreps_in2 = irreps_in2
        self.irreps_out = irreps_out
        self.weight_numel = self.stride_w

        self.internal_weights = False
        self.shared_weights = shared_weights
        if internal_weights is not None:
            self.internal_weights = True
            self.weight = nn.Parameter(internal_weights)

        if self.e3nn_compat_mode:
            _irreps_in1_segment_info = _irreps_to_segment_info(self.irreps_in1)
            _irreps_in2_segment_info = _irreps_to_segment_info(self.irreps_in2)
            _irreps_out_segment_info = _irreps_to_segment_info(self.irreps_out)

            self.register_buffer("_irreps_in1_segment_info", _irreps_in1_segment_info)
            self.register_buffer("_irreps_in2_segment_info", _irreps_in2_segment_info)
            self.register_buffer("_irreps_out_segment_info", _irreps_out_segment_info)

        self._tp_info_fwd = TensorProductPathData(*tp_infos[0])
        self._tp_info_bwd_dgrad_in1 = TensorProductPathData(*tp_infos[1])
        self._tp_info_bwd_dgrad_in2 = TensorProductPathData(*tp_infos[2])
        self._tp_info_bwd_dgrad_w = TensorProductPathData(*tp_infos[3])

    def forward(
        self,
        in1: torch.Tensor,
        in2: torch.Tensor,
        weight: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        args = (
            self._tp_info_fwd.c_type(),
            self._tp_info_bwd_dgrad_in1.c_type(),
            self._tp_info_bwd_dgrad_in2.c_type(),
            self._tp_info_bwd_dgrad_w.c_type(),
            self.connection_mode,
            self.stride_out,
        )
        if self.e3nn_compat_mode:
            in1 = transpose_irrep_to_channels_last(in1, self._irreps_in1_segment_info)
            in2 = transpose_irrep_to_channels_last(in2, self._irreps_in2_segment_info)

        # TODO: better checks against shape matches and other things
        if self.internal_weights:
            weight = self.weight
        out = _fused_tensor_product_autograd.apply(in1, in2, weight, None, args)

        if self.e3nn_compat_mode:
            out = transpose_irrep_to_m_last(out, self._irreps_out_segment_info)

        return out


class FusedFullyConnectedTensorProduct(FusedTensorProduct):
    def __init__(
        self,
        irreps_in1: o3.Irreps,
        irreps_in2: o3.Irreps,
        irreps_out: o3.Irreps,
        connection_mode: str = "uvw",
        irrep_normalization: Optional[str] = None,
        path_normalization: Optional[str] = None,
        e3nn_compat_mode: bool = False,
        internal_weights: Optional[bool] = None,
        shared_weights: Optional[bool] = None,
    ):
        connection_mode = connection_mode.lower()
        supported_modes = {"uvw", "uvu", "uvv", "uuw"}
        if connection_mode not in supported_modes:
            msg = f"Invalid connection_mode, received {connection_mode}"
            msg += f", supported modes are {supported_modes}."
            raise ValueError(msg)

        instr = [
            (i_1, i_2, i_out, connection_mode, True, 1.0)
            for i_1, (_, ir_1) in enumerate(irreps_in1)
            for i_2, (_, ir_2) in enumerate(irreps_in2)
            for i_out, (_, ir_out) in enumerate(irreps_out)
            if ir_out in ir_1 * ir_2
        ]

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            _tp = o3.TensorProduct(
                irreps_in1,
                irreps_in2,
                irreps_out,
                instr,
                internal_weights=internal_weights,
                shared_weights=shared_weights,
                irrep_normalization=irrep_normalization,
                path_normalization=path_normalization,
            )

        super().__init__(
            _tp.instructions,
            _tp.irreps_in1,
            _tp.irreps_in2,
            _tp.irreps_out,
            e3nn_compat_mode=e3nn_compat_mode,
            internal_weights=_tp.weight.data if internal_weights else None,
            shared_weights=shared_weights,
        )


class FusedFullTensorProduct(FusedTensorProduct):
    def __init__(
        self,
        irreps_in1: o3.Irreps,
        irreps_in2: o3.Irreps,
        filter_ir_out: Iterator[o3.Irrep] = None,
        irrep_normalization: str = None,
        e3nn_compat_mode: bool = False,
    ):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            _tp = o3.FullTensorProduct(
                irreps_in1,
                irreps_in2,
                filter_ir_out=filter_ir_out,
                irrep_normalization=irrep_normalization,
            )

        super().__init__(
            _tp.instructions,
            _tp.irreps_in1,
            _tp.irreps_in2,
            _tp.irreps_out,
            e3nn_compat_mode=e3nn_compat_mode,
        )


class FusedElementWiseTensorProduct(FusedTensorProduct):
    def __init__(
        self,
        irreps_in1: o3.Irreps,
        irreps_in2: o3.Irrep,
        filter_ir_out=None,
        irrep_normalization: str = None,
        e3nn_compat_mode: bool = False,
    ):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            _tp = o3.ElementwiseTensorProduct(
                irreps_in1,
                irreps_in2,
                filter_ir_out=filter_ir_out,
                irrep_normalization=irrep_normalization,
            )

        super().__init__(
            _tp.instructions,
            _tp.irreps_in1,
            _tp.irreps_in2,
            _tp.irreps_out,
            e3nn_compat_mode=e3nn_compat_mode,
        )


class FusedLinear(nn.Module):
    def __init__(
        self,
        irreps_in: o3.Irreps,
        irreps_out: o3.Irreps,
        e3nn_compat_mode: bool = False,
        internal_weights: Optional[bool] = None,
        shared_weights: Optional[bool] = None,
    ):
        super().__init__()

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            _tp = o3.Linear(
                irreps_in,
                irreps_out,
                internal_weights=internal_weights,
                shared_weights=shared_weights,
            )

        self.internal_weights = internal_weights
        self.shared_weights = shared_weights
        if internal_weights:
            self.weight = nn.Parameter(_tp.weight.data)

        tp_infos = get_linear_infos_from_e3nn_instructions(
            _tp.instructions,
            irreps_in,
            irreps_out,
        )

        self.e3nn_compat_mode = e3nn_compat_mode
        self.connection_mode = tp_infos[3]
        self.stride_in = tp_infos[4][0]
        self.stride_w = tp_infos[4][1]
        self.stride_out = tp_infos[4][3]
        self.irreps_in = irreps_in
        self.irreps_out = irreps_out
        self.weight_numel = self.stride_w

        if self.e3nn_compat_mode:
            _irreps_in_segment_info = _irreps_to_segment_info(self.irreps_in)
            _irreps_out_segment_info = _irreps_to_segment_info(self.irreps_out)

            self.register_buffer("_irreps_in_segment_info", _irreps_in_segment_info)
            self.register_buffer("_irreps_out_segment_info", _irreps_out_segment_info)

        self._tp_info_fwd = TensorProductPathData(*tp_infos[0])
        self._tp_info_bwd_dgrad_in = TensorProductPathData(*tp_infos[1])
        self._tp_info_bwd_dgrad_w = TensorProductPathData(*tp_infos[2])

    def forward(self, x: torch.Tensor, weight: Optional[torch.Tensor] = None):
        args = (
            self._tp_info_fwd.c_type(),
            self._tp_info_bwd_dgrad_in.c_type(),
            self._tp_info_bwd_dgrad_w.c_type(),
            self._tp_info_bwd_dgrad_w.c_type(),
            self.connection_mode,
            self.stride_out,
        )

        if self.e3nn_compat_mode:
            x = transpose_irrep_to_channels_last(x, self._irreps_in_segment_info)

        # TODO: better checks
        if self.internal_weights:
            weight = self.weight

        out = _fused_tensor_product_autograd.apply(x, weight, None, None, args)

        if self.e3nn_compat_mode:
            out = transpose_irrep_to_m_last(out, self._irreps_out_segment_info)

        return out


class FusedMainContraction(nn.Module):
    def __init__(
        self,
        tp_infos: List[Tuple[torch.Tensor]],
        strides: List[int],
    ):
        super().__init__()

        self._tp_info_fwd = TensorProductPathData(*tp_infos[0])
        self._tp_info_bwd_dgrad_x = TensorProductPathData(*tp_infos[1])
        self._tp_info_bwd_dgrad_y = TensorProductPathData(*tp_infos[2])
        self._tp_info_bwd_dgrad_w = TensorProductPathData(*tp_infos[3])

        self.connection_mode = getattr(pylibcugraphops.ConnectionMode, "UVU")

        self.stride_x = strides[0]
        self.stride_y = strides[1]
        self.stride_w = strides[2]
        self.stride_o = strides[3]

    def forward(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        contracted_weight: torch.Tensor,
        next_contracted_weight: Optional[torch.Tensor],
    ):
        args = (
            self._tp_info_fwd.c_type(),
            self._tp_info_bwd_dgrad_x.c_type(),
            self._tp_info_bwd_dgrad_y.c_type(),
            self._tp_info_bwd_dgrad_w.c_type(),
            self.connection_mode,
            self.stride_o,
        )

        out = _fused_tensor_product_autograd.apply(
            x, y, contracted_weight, next_contracted_weight, args
        )

        return out


class FusedFeatureContraction(nn.Module):
    def __init__(
        self,
        tp_infos: List[Tuple[torch.Tensor]],
        strides: List[int],
    ):
        super().__init__()

        self._tp_info_fwd = TensorProductPathData(*tp_infos[0])
        self._tp_info_bwd_dgrad_x = TensorProductPathData(*tp_infos[1])
        self._tp_info_bwd_dgrad_w = TensorProductPathData(*tp_infos[2])

        self.connection_mode = getattr(pylibcugraphops.ConnectionMode, "UUU")

        self.stride_x = strides[0]
        self.stride_w = strides[1]
        self.stride_o = strides[3]

    def forward(
        self,
        x: torch.Tensor,
        contracted_weight: torch.Tensor,
        next_contracted_weight: Optional[torch.Tensor],
    ):
        args = (
            self._tp_info_fwd.c_type(),
            self._tp_info_bwd_dgrad_x.c_type(),
            self._tp_info_bwd_dgrad_w.c_type(),
            # duplication only for API compatibility until fixed, not accessed
            self._tp_info_bwd_dgrad_w.c_type(),
            self.connection_mode,
            self.stride_o,
        )

        out = _fused_tensor_product_autograd.apply(
            x, contracted_weight, None, next_contracted_weight, args
        )

        return out


class FusedSymmetricContraction(nn.Module):
    def __init__(
        self,
        irreps_in: o3.Irreps,
        irreps_out: o3.Irreps,
        correlation: int,
        num_elements: int,
        e3nn_compat_mode: bool = False,
    ):
        super().__init__()

        self.irreps_in = irreps_in
        self.irreps_out = irreps_out
        self.correlation = correlation
        self.num_irreps_out = len(irreps_out)
        self.num_elements = num_elements
        self.num_features = irreps_in.count((0, 1))
        self.e3nn_compat_mode = e3nn_compat_mode

        if self.e3nn_compat_mode:
            _irreps_out_segment_info = _irreps_to_segment_info(self.irreps_out)
            self.register_buffer("_irreps_out_segment_info", _irreps_out_segment_info)

        u_matrices = get_sym_cont_u_matrices(irreps_in, irreps_out, correlation)
        weights = init_sym_cont_weights(
            u_matrices, self.num_elements, self.num_features, correlation
        )

        weight_dict = {}
        for idx in range(self.num_irreps_out):
            for nu in range(0, correlation):
                self.register_buffer(f"u_matrix_{idx}_{nu}", u_matrices[idx][nu])
                weight_dict[f"weight_{idx}_{nu}"] = weights[idx][nu]
        self.weights = nn.ParameterDict(weight_dict)

        tp_infos, weight_indices, strides = get_sym_contraction_path_offsets(
            weights, u_matrices, correlation, self.num_features, self.num_elements
        )

        for nu in range(correlation):
            for i, idx in enumerate(weight_indices[nu]):
                self.register_buffer(f"weight_selection_index_{nu}_{i}", idx)
        self._idx_len = [len(idx) for idx in weight_indices]

        self.main_contraction = FusedMainContraction(
            tp_infos[correlation - 1], strides[correlation - 1]
        )
        feature_contractions = []
        for nu in range(0, correlation - 1):
            feature_contractions.append(
                FusedFeatureContraction(tp_infos[nu], strides[nu])
            )
        self.feature_contractions = nn.ModuleList(feature_contractions)

    def forward(self, x: torch.Tensor, y: torch.Tensor):
        if self.e3nn_compat_mode:
            # if compat mode, x is of shape [BS, NUM_FEAT, NUM_ELL]
            # we need [BS, NUM_ELL*NUM_FEAT]
            x = x.permute(0, 2, 1).contiguous()
            x = x.view(x.size(0), -1)

        u_matrices = self._get_u_matrices_from_buffers()
        weights = self._get_weights_from_buffers()
        contracted_weights = contract_sym_cont_weights(
            weights, u_matrices, self.correlation
        )

        out = None

        if self.correlation == 1:
            cw = self._index_contracted_weights(
                self.correlation - 1, contracted_weights
            )
            cw = cw.view(-1)
            out = self.main_contraction(x, y, cw, None)

        else:
            cw = self._index_contracted_weights(
                self.correlation - 1, contracted_weights
            )
            cw = cw.view(-1)
            cw_next = self._index_contracted_weights(
                self.correlation - 2, contracted_weights
            )
            cw_next = self._contract_weight(cw_next, y)
            out = self.main_contraction(x, y, cw, cw_next)

        for nu in range(self.correlation - 2, -1, -1):
            cw_next = None
            if nu - 1 >= 0:
                cw_next = self._index_contracted_weights(nu - 1, contracted_weights)
                cw_next = self._contract_weight(cw_next, y)
            out = self.feature_contractions[nu](x, out, cw_next)

        if self.e3nn_compat_mode:
            out = transpose_irrep_to_m_last(out, self._irreps_out_segment_info)

        return out

    def _contract_weight(self, cw: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        cw = cw.view(-1, cw.size(-1))
        c_tensor = torch.einsum("be,ae->ba", y, cw)
        return c_tensor

    def _index_contracted_weights(
        self, nu: int, contracted_weights: List[torch.Tensor]
    ) -> torch.Tensor:
        idx = self._get_weight_selection_indices(nu)
        cw = contracted_weights[nu][idx]
        return cw

    def _get_u_matrices_from_buffers(self):
        u_matrices = []
        for idx in range(self.num_irreps_out):
            tmp = []
            for nu in range(0, self.correlation):
                tmp.append(getattr(self, f"u_matrix_{idx}_{nu}"))
            u_matrices.append(tmp)
        return u_matrices

    def _get_weights_from_buffers(self):
        weights = []
        for idx in range(self.num_irreps_out):
            tmp = []
            for nu in range(0, self.correlation):
                tmp.append(self.weights[f"weight_{idx}_{nu}"])
            weights.append(tmp)
        return weights

    def _load_weights_from_reference(self, module: nn.Module):
        for i in range(len(self.irreps_out)):
            ref_layer = module.contractions[i]
            for nu in range(self.correlation):
                if nu == (self.correlation - 1):
                    weight = ref_layer.weights_max
                else:
                    weight = ref_layer.weights[self.correlation - 2 - nu]

                self.weights[f"weight_{i}_{nu}"].data = weight.data.clone()

    def _get_weight_selection_indices(self, nu: int) -> Tuple:
        return tuple(
            [
                getattr(self, f"weight_selection_index_{nu}_{i}")
                for i in range(self._idx_len[nu])
            ]
        )


class LoopedFusedSymmetricContraction(nn.Module):
    def __init__(
        self,
        irreps_in: o3.Irreps,
        irreps_out: o3.Irreps,
        correlation: Union[int, Iterator[int]],
        num_elements: int,
        e3nn_compat_mode: bool = False,
    ):
        # unlike the original implementation, FusedContraction
        # already would work for many different irreps at once
        # however, the current implementation for now requires rather
        # a lot of shared memory and to be compatible with having
        # different correlations for different irreps, this can be
        # used instead

        super().__init__()
        self.irreps_in = irreps_in
        self.irreps_out = irreps_out
        self.correlation = correlation

        if isinstance(correlation, int):
            correlation = [correlation] * len(irreps_out)
        else:
            if len(correlation) != len(irreps_out):
                raise ValueError("List of iteration values does not match irreps_out.")

        self.contractions = torch.nn.ModuleList()
        for i, irrep_out in enumerate(self.irreps_out):
            self.contractions.append(
                FusedSymmetricContraction(
                    irreps_in,
                    o3.Irreps(str(irrep_out)),
                    correlation=correlation[i],
                    num_elements=num_elements,
                    e3nn_compat_mode=e3nn_compat_mode,
                )
            )

    def forward(self, x: torch.Tensor, y: torch.Tensor):
        outs = [contraction(x, y) for contraction in self.contractions]
        return torch.cat(outs, dim=1)

    def _load_weights_from_reference(self, module: nn.Module):
        for i in range(len(self.irreps_out)):
            ref_layer = module.contractions[i]
            for nu in range(self.correlation):
                if nu == (self.correlation - 1):
                    weight = ref_layer.weights_max
                else:
                    weight = ref_layer.weights[self.correlation - 2 - nu]

                self.contractions[i].weights[
                    f"weight_{0}_{nu}"
                ].data = weight.data.clone()



# Copyright (c) 2022-2024, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import pylibcugraphops
import warnings

from typing import Optional, Union, Tuple


class CSC:
    def __init__(
        self,
        offsets: torch.Tensor,
        indices: torch.Tensor,
        num_src_nodes: int,
        map_csc_to_coo: Optional[torch.Tensor] = None,
        map_dst_to_src: Optional[torch.Tensor] = None,
        dst_max_in_degree: int = -1,
        reverse_graph_bwd: bool = False,
        is_bipartite: bool = False,
    ):
        self.offsets = offsets
        self.indices = indices
        self.num_src_nodes = num_src_nodes
        self.num_dst_nodes = offsets.size(0) - 1
        self.num_indices = indices.size(0)

        self.rev_offsets = (
            torch.empty(
                num_src_nodes + 1,
                device=offsets.device,
                dtype=offsets.dtype,
            )
            if reverse_graph_bwd
            else None
        )
        self.rev_indices = torch.empty_like(indices) if reverse_graph_bwd else None

        self.map_dst_to_src = map_dst_to_src
        self.map_csc_to_coo = map_csc_to_coo
        self.map_rev_to_coo = torch.empty_like(indices) if reverse_graph_bwd else None

        self.dst_max_in_degree = dst_max_in_degree
        # TODO remove warning once we again have support for these
        if dst_max_in_degree > 0:
            warnings.warn("dst_max_in_degree currently has no effect")
        if map_dst_to_src is not None:
            raise NotImplementedError("map_dst_to_src is currently not supported")

        self.is_bipartite = is_bipartite

        if reverse_graph_bwd:
            self.add_reverse_graph()

        else:
            if self.is_bipartite:
                self.graph_csc = pylibcugraphops.make_bipartite_csc(
                    offsets,
                    indices,
                    num_src_nodes,
                    map_csc_to_coo,
                )

            else:
                self.graph_csc = pylibcugraphops.make_csc(
                    offsets,
                    indices,
                    num_src_nodes,
                    map_csc_to_coo,
                )

        self.reverse_graph_bwd = reverse_graph_bwd

    def __to__(self, *args, **kwargs):
        self.offsets = self.offsets.to(*args, **kwargs)
        self.indices = self.indices.to(*args, **kwargs)
        if self.map_dst_to_src is not None:
            self.map_dst_to_src = self.map_dst_to_src.to(*args, **kwargs)
        if self.map_csc_to_coo is not None:
            self.map_csc_to_coo = self.map_csc_to_coo.to(*args, **kwargs)
        if self.map_rev_to_coo is not None:
            self.map_rev_to_coo = self.map_rev_to_coo.to(*args, **kwargs)
        if self.reverse_graph_bwd:
            self.rev_offsets = self.rev_offsets.to(*args, **kwargs)
            self.rev_indices = self.rev_indices.to(*args, **kwargs)

    def add_reverse_graph(self):
        if self.reverse_graph_bwd:
            assert self.rev_offsets is not None
            assert self.rev_indices is not None
            assert self.map_rev_to_coo is not None
            return

        self.rev_offsets = torch.empty(
            self.num_src_nodes + 1,
            device=self.offsets.device,
            dtype=self.offsets.dtype,
        )

        self.rev_indices = torch.empty_like(self.indices)
        self.map_rev_to_coo = torch.empty_like(self.indices)

        if self.is_bipartite:
            self.graph_csc = pylibcugraphops.make_bipartite_csc(
                self.offsets,
                self.indices,
                self.num_src_nodes,
                self.map_csc_to_coo,
                self.rev_offsets,
                self.rev_indices,
                self.map_rev_to_coo,
            )

        else:
            self.graph_csc = pylibcugraphops.make_csc(
                self.offsets,
                self.indices,
                self.num_src_nodes,
                self.map_csc_to_coo,
                self.rev_offsets,
                self.rev_indices,
                self.map_rev_to_coo,
            )

        stream = torch.cuda.current_stream().cuda_stream
        # call reverse to get necessary workspace size and
        # create necessary workspace buffers
        node_counts = torch.empty(
            self.num_src_nodes,
            dtype=self.offsets.dtype,
            device=self.offsets.device,
        )
        cub_workspace = None
        cub_workspace_size = pylibcugraphops.reverse_graph(
            self.graph_csc, node_counts, cub_workspace, 0, stream
        )
        cub_workspace = torch.empty(
            cub_workspace_size, dtype=torch.int8, device=self.offsets.device
        )
        # actually reverse graph (in-place)
        pylibcugraphops.reverse_graph(
            self.graph_csc, node_counts, cub_workspace, cub_workspace_size, stream
        )


class HeteroCSC(CSC):
    def __init__(
        self,
        offsets: torch.Tensor,
        indices: torch.Tensor,
        edge_types: torch.Tensor,
        num_src_nodes: int,
        num_edge_types: int,
        num_node_types: int = 0,
        node_types: Optional[Union[torch.Tensor, Tuple[torch.Tensor]]] = None,
        map_dst_to_src: Optional[torch.Tensor] = None,
        map_csc_to_coo: Optional[torch.Tensor] = None,
        dst_max_in_degree: int = -1,
        reverse_graph_bwd: bool = False,
        is_bipartite: bool = False,
    ):
        if is_bipartite:
            if num_node_types > 0:
                assert isinstance(node_types, Tuple), (
                    "Expected tuple of node types for a bipartite graph"
                    f", but got {type(node_types)} instead."
                )

        else:
            if num_node_types > 0:
                assert isinstance(node_types, torch.Tensor), (
                    "Expected node_types to be of type torch.Tensor"
                    f", but got {type(node_types)} instead."
                )

        super(HeteroCSC, self).__init__(
            offsets,
            indices,
            num_src_nodes,
            map_dst_to_src=map_dst_to_src,
            map_csc_to_coo=map_csc_to_coo,
            dst_max_in_degree=dst_max_in_degree,
            reverse_graph_bwd=reverse_graph_bwd,
            is_bipartite=is_bipartite,
        )
        self.edge_types = edge_types
        self.num_edge_types = num_edge_types

        if self.is_bipartite:
            src_node_types, dst_node_types = node_types
            self.graph_csc = pylibcugraphops.make_bipartite_csc_hg(
                offsets,
                indices,
                num_src_nodes,
                num_node_types,
                num_edge_types,
                src_node_types,
                dst_node_types,
                edge_types,
                map_csc_to_coo,
                self.rev_offsets,
                self.rev_indices,
                self.map_rev_to_coo,
            )

        else:
            self.graph_csc = pylibcugraphops.make_csc_hg(
                offsets,
                indices,
                num_src_nodes,
                num_node_types,
                num_edge_types,
                node_types,
                edge_types,
                map_csc_to_coo,
                self.rev_offsets,
                self.rev_indices,
                self.map_rev_to_coo,
            )

    def __to__(self, *args, **kwargs):
        super(HeteroCSC, self).__to__(*args, **kwargs)
        self.edge_types = self.edge_types.to(*args, **kwargs)


#################################################################################
# Deprecated Graph Types
# for now, they inherit from CSC which mostly makes all previous types
# (in 23.04) compatible with the refactored operators. Deprecation in 23.08,
# potential deletion in 23.10.
#################################################################################


class BipartiteCSC(CSC):
    def __init__(
        self,
        offsets: torch.Tensor,
        indices: torch.Tensor,
        num_src_nodes: int,
        ef_indices: Optional[torch.Tensor] = None,
        reverse_graph_bwd: bool = False,
    ):
        super(BipartiteCSC, self).__init__(
            offsets,
            indices,
            num_src_nodes=num_src_nodes,
            map_csc_to_coo=ef_indices,
            reverse_graph_bwd=reverse_graph_bwd,
            is_bipartite=True,
        )

        warnings.warn(
            "BipartiteCSC is deprecated with the 23.08 release"
            " and will be removed in 23.10, use CSC instead.",
            DeprecationWarning,
        )


class SampledCSC(CSC):
    def __init__(
        self,
        offsets: torch.Tensor,
        indices: torch.Tensor,
        max_num_neighbors: int,
        num_src_nodes: int,
    ):
        super(SampledCSC, self).__init__(
            offsets,
            indices,
            num_src_nodes=num_src_nodes,
            dst_max_in_degree=max_num_neighbors,
            is_bipartite=False,
        )

        warnings.warn(
            "SampledCSC is deprecated with the 23.08 release"
            " and will be removed in 23.10, use CSC instead.",
            DeprecationWarning,
        )


class StaticCSC(CSC):
    def __init__(
        self,
        offsets: torch.Tensor,
        indices: torch.Tensor,
        ef_indices: Optional[torch.Tensor] = None,
        reverse_graph_bwd: bool = False,
    ):
        num_src_nodes = offsets.size(0) - 1
        super(StaticCSC, self).__init__(
            offsets,
            indices,
            num_src_nodes=num_src_nodes,
            map_csc_to_coo=ef_indices,
            reverse_graph_bwd=reverse_graph_bwd,
            is_bipartite=False,
        )

        warnings.warn(
            "StaticCSC is deprecated with the 23.08 release"
            " and will be removed in 23.10, use CSC instead.",
            DeprecationWarning,
        )


class SampledHeteroCSC(HeteroCSC):
    def __init__(
        self,
        offsets: torch.Tensor,
        indices: torch.Tensor,
        edge_types: torch.Tensor,
        max_num_neighbors: int,
        num_src_nodes: int,
        num_edge_types: int,
    ):
        super(SampledHeteroCSC, self).__init__(
            offsets,
            indices,
            edge_types=edge_types,
            num_src_nodes=num_src_nodes,
            num_edge_types=num_edge_types,
            dst_max_in_degree=max_num_neighbors,
            is_bipartite=False,
        )

        warnings.warn(
            "SampledHeteroCSC is deprecated with the 23.08 release"
            " and will be removed in 23.10, use HeteroCSC instead.",
            DeprecationWarning,
        )


class StaticHeteroCSC(HeteroCSC):
    def __init__(
        self,
        offsets: torch.Tensor,
        indices: torch.Tensor,
        edge_types: torch.Tensor,
        num_edge_types: int,
        ef_indices: Optional[torch.Tensor] = None,
    ):
        num_src_nodes = offsets.size(0) - 1

        super(StaticHeteroCSC, self).__init__(
            offsets,
            indices,
            edge_types=edge_types,
            num_src_nodes=num_src_nodes,
            num_edge_types=num_edge_types,
            map_csc_to_coo=ef_indices,
            is_bipartite=False,
        )

        warnings.warn(
            "StaticHeteroCSC is deprecated with the 23.08 release"
            " and will be removed in 23.10, use HeteroCSC instead.",
            DeprecationWarning,
        )
